
package mapimporter;

/**
 *
 * @author David Fernandez
 */

import java.io.*;


public class CreateMap {

    static int MAX_PLAYERS = 64;

    public static int Y(int y, int gridy) {
        // Civ4 uses "y" coordinate from South to North, while in freeciv "Y" increases from North to South
        return (gridy-1-y);
    }

    public static void Create(String inCiv4, String inFreeciv, String outFreeciv, int riverOffset, String[] scenarioNames) throws Exception {
        String linea;
        String[] campo;
        int i = 0;

        int gridx = 0, gridy = 0;
        int x = 0, y = 0;
        int team = 0;
        boolean isRiver = false;
        boolean isNofRiver = false;
        boolean isWofRiver = false;
        String featureType = "no";
        String terrainType = "no";
        String plotType = "no";

        char terrain;
        char terrainMap[][];
        boolean riverMap[][];
        int startingPos[][];

        int fixedX = 0;
        int fixedY = 0;

        try {
            LineNumberReader inMap = new LineNumberReader(new FileReader(inCiv4));           // Input Civ4 File
            LineNumberReader inScenario = new LineNumberReader(new FileReader(inFreeciv));   // Input Freeciv File
            PrintWriter outMap = new PrintWriter(new FileWriter(outFreeciv));                  // Output File

            startingPos = new int[MAX_PLAYERS][2];
            for (i=0; i<MAX_PLAYERS; i++) {
                startingPos[i][0] = -1;
                startingPos[i][1] = -1;
            }
            // Read starting positions placed under "BeginPlayer"
            while ( ((linea = inMap.readLine()) != null ) && (linea.startsWith("BeginMap") == false) ) {
                linea=linea.trim();
                if (linea.startsWith("Team=")) {
                    team = Integer.valueOf(linea.substring(linea.indexOf("=")+1));
                }
                if (linea.startsWith("StartingX=")) {
                    campo = linea.split(",");
                    if (team < MAX_PLAYERS) {
                        campo[0] = campo[0].substring(linea.indexOf("=")+1);
                        startingPos[team][0] = Integer.valueOf(campo[0].trim());
                        campo[1] = campo[1].substring(linea.indexOf("=")+2);
                        startingPos[team][1] = Integer.valueOf(campo[1].trim());
                    }
                }
            }

            // Read map size placed under "BeginMap"
            while ( ((linea = inMap.readLine()) != null ) && (linea.startsWith("BeginPlot") == false) ) {
                linea=linea.trim();
                if (linea.startsWith("grid width=")) {
                    gridx = Integer.valueOf(linea.substring(linea.indexOf("=")+1));
                }
                if (linea.startsWith("grid height=")) {
                    gridy = Integer.valueOf(linea.substring(linea.indexOf("=")+1));
                }
            }
            if ( (gridx == 0) || (gridy == 0)) {
                throw new Exception (inCiv4 + " does not appear to be a valid Civ4 Map.");
            }

            terrainMap = new char[gridx][gridy];
            riverMap = new boolean[gridx][gridy];
            for (y=0; y<gridy; y++) {
                for (x=0; x<gridx; x++) {
                    terrainMap [x][y] = '+';
                    riverMap [x][y] = false;
                }
            }
            // Read terrain types and rivers placed under "BeginPlot"
            while ( ((linea = inMap.readLine()) != null ) ) {
                linea=linea.trim();
                if (linea.startsWith("x=")) {
                    campo = linea.split(",");
                    x = Integer.valueOf(campo[0].substring(linea.indexOf("=")+1));
                    y = Integer.valueOf(campo[1].substring(linea.indexOf("=")+1));//
                }
                if (linea.startsWith("isNOfRiver")) {
                    isRiver = true;
                    isNofRiver = true;
                }
                if (linea.startsWith("isWOfRiver")) {
                    isRiver = true;
                    isWofRiver = true;
                }
                if (linea.startsWith("FeatureType=")) {
                    campo = linea.split(",");
                    featureType = campo[0].substring(linea.indexOf("=")+1);
                }
                if (linea.startsWith("TerrainType=")) {
                    campo = linea.split(",");
                    terrainType = campo[0].substring(linea.indexOf("=")+1);
                }
                if (linea.startsWith("PlotType=")) {
                    campo = linea.split(",");
                    plotType = campo[0].substring(linea.indexOf("=")+1);
                }
                // Link each civ4 terrain type with freeciv char identifiers (rare lakes, and rare swamps)
                if (linea.startsWith("EndPlot")) {
                    if (plotType.equals("0")) terrain = 'm';
                    else if (plotType.equals("1")) terrain = 'h';
                    else if (plotType.equals("2") && terrainType.equals("TERRAIN_COAST")) terrain = '+';
                    else if (plotType.equals("3") && isRiver) terrain = 's';
                    else if (featureType.equals("FEATURE_ICE")) terrain = 'a';
                    else if (terrainType.equals("TERRAIN_SNOW")) terrain = 'a';
                    else if (terrainType.equals("TERRAIN_OCEAN")) terrain = ':';
                    else if (terrainType.equals("TERRAIN_COAST")) terrain = ' ';
                    else if (terrainType.equals("TERRAIN_DESERT")) terrain = 'd';
                    else if (featureType.equals("FEATURE_FOREST")) terrain = 'f';
                    else if (featureType.equals("FEATURE_JUNGLE")) terrain = 'j';
                    else if (terrainType.equals("TERRAIN_PLAINS")) terrain = 'p';
                    else if (terrainType.equals("TERRAIN_GRASS")) terrain = 'g';
                    else if (terrainType.equals("TERRAIN_TUNDRA")) terrain = 't';
                    else terrain = '+';

                    if ( (x < gridx) && (y >= 0) ) {
                        terrainMap[x][Y(y,gridy)] = terrain;

                        // Rivers placed in tile joints, adjusted to the center of the N-W tile by default
                        if (isRiver) {

                            switch (riverOffset) {
                                case 4: {   //S-E
                                    fixedX = (x+1)%gridx;
                                    fixedY = Y(y,gridy)+1;
                                    break;
                                }
                                case 3: {   //S-W
                                    fixedX = x;
                                    fixedY = Y(y,gridy)+1;
                                    break;
                                }
                                case 2: {   //N-E
                                    fixedX = (x+1)%gridx;
                                    fixedY = Y(y,gridy);
                                    break;
                                }
                                default: {   //N-W
                                    fixedX = x;
                                    fixedY = Y(y,gridy);
                                    break;
                                }
                            }
                            if (fixedY>=0 && fixedY<gridy) {
                                riverMap[fixedX][fixedY] = true;
                                if (isNofRiver) {
                                    // Needed to keep the same topology of the river while flowing through the tiles than it was in the joints
                                    riverMap[(fixedX-1)%gridx][fixedY] = true;
                                }
                            }
                        }
                    }

                    // Reset values for next loop, in case some entry is missing
                    x = 0; y = 0;
                    featureType = "no";
                    terrainType = "no";
                    plotType = "no";
                    isRiver = false;
                    isNofRiver = false;
                    isWofRiver = false;
                }
            }

            // Read scenario options from Freeciv Scenario input
            while ( ((linea = inScenario.readLine()) != null ) && (linea.startsWith("[map]") == false) ) {

                if (linea.startsWith("name=_(\"")) {
                    outMap.println("name=_(\"" + scenarioNames[0] + "\")");
                }
                else if (linea.startsWith("description=_(\"")) {
                    outMap.println("description=_(\"" + scenarioNames[1] + "\")");
                }
                else if (linea.startsWith("metastring=\"")) {
                    outMap.println("metastring=\"" + scenarioNames[2] + "\"");
                }
                else outMap.println(linea);
            }
            if (linea == null) {
                throw new Exception (inFreeciv + " does not appear to be a valid Freeciv Scenario.");
            }

            outMap.printf("[map]");

            // Write the map sizes
            outMap.printf("\nwidth=%d", gridx);
            outMap.printf("\nheight=%d", gridy);

            // Write the other map options
            outMap.printf("\nseed=0");
            outMap.printf("\nlandpercent=33");
            outMap.printf("\nriches=250");
            outMap.printf("\nswampsize=1");
            outMap.printf("\ndeserts=60");
            outMap.printf("\nriverlength=259");
            outMap.printf("\nmountains=51");
            outMap.printf("\nforestsize=159");
            outMap.printf("\nhuts=50");
            outMap.printf("\ngenerator=0");
            outMap.printf("\nis_earth=1");

            // Write starting positions
            for (team=0; (team<MAX_PLAYERS) && (startingPos[team][0]>=0); team++) {
                outMap.printf( "\nr%dsx=%d", team, startingPos[team][0]);
                outMap.printf( "\nr%dsy=%d", team, Y(startingPos[team][1],gridy));
            }

            // Write terrain types
            for (y=0; y<gridy; y++) {
                outMap.printf( "\nt%03d=\"", y);
                for (x=0; x<gridx; x++) {
                    outMap.print(terrainMap[x][y]);
                }
                outMap.printf("\"");
            }

            // Write rivers as overlays
            for (y=0; y<gridy; y++) {
                outMap.printf( "\nn%03d=\"", y);
                for (x=0; x<gridx; x++) {
                    if ((riverMap[x][y]) &&
                            (terrainMap[x][y] != ' ') &&    //no deep ocean
                            (terrainMap[x][y] != ':') &&    //no coast ocean
                            (terrainMap[x][y] != '+') )     //no lake
                    {
                        outMap.print("2");
                    } else {
                        outMap.print("0");
                    }
                }
                outMap.printf("\"");
            }

            inMap.close();
            inScenario.close();
            outMap.close();
            System.out.println("Successfully saved to: " + outFreeciv + ".");
        } catch (Exception e) {
            System.err.println(e.getLocalizedMessage());
        }
    }
}